#include <stdlib.h>
#include <unistd.h>
#include <ctpublic.h>
#include <ctlib.h>

static char  software_version[]   = "$Id: ct.c,v 1.11 1999/01/23 19:28:49 camber Exp $";
static void *no_unused_var_warn[] = {software_version,
                                     no_unused_var_warn};


int (*g_ctlib_msg_handler)() = NULL;
int (*g_ctlib_err_handler)() = NULL;
extern int (*g_tds_msg_handler)();
extern int (*g_tds_err_handler)();

static void _ct_bind_data(CS_COMMAND *cmd);


CS_RETCODE ct_exit(CS_CONTEXT *ctx, CS_INT unused)
{
	return CS_SUCCEED;
}
CS_RETCODE ct_init(CS_CONTEXT *ctx, CS_INT version)
{
   /* set the functions in the TDS layer to point to the correct info/err
    * message handler functions */
   g_tds_msg_handler = ctlib_handle_info_message;
   g_tds_err_handler = ctlib_handle_err_message;
	return CS_SUCCEED;
}
CS_RETCODE ct_con_alloc(CS_CONTEXT *ctx, CS_CONNECTION **con)
{
	*con = (CS_CONNECTION *) malloc(sizeof(CS_CONNECTION));
	(*con)->tds_login = (void *) tds_alloc_login();

	/* so we know who we belong to */
	(*con)->ctx = ctx;

	/* set default values */
	tds_set_library((*con)->tds_login, "CT-Library");
	tds_set_charset((*con)->tds_login, "iso_1");
	tds_set_packet((*con)->tds_login, 512);
	return CS_SUCCEED;
}
CS_RETCODE ct_callback()
{
	return CS_SUCCEED;
}
CS_RETCODE ct_con_props(CS_CONNECTION *con, CS_INT action, CS_INT property,
CS_VOID *buffer, CS_INT buflen, CS_INT *out_len)
{
	/* FIX ME, this will break if buflen != CS_NULLTERM */
	if (action==CS_SET) {
		switch (property) {
			case CS_USERNAME:
				tds_set_user(con->tds_login,(char *)buffer);
				break;
			case CS_PASSWORD:
				tds_set_passwd(con->tds_login,(char *)buffer);
				break;
			default:
				fprintf(stderr,"Unknown property %d\n",property);
		}
	} /* else if */
	return CS_SUCCEED;
}
CS_RETCODE ct_connect(CS_CONNECTION *con, CS_CHAR *servername, CS_INT snamelen)
{
char *server;
int needfree=0;

	if (snamelen==0 || snamelen==CS_UNUSED) {
		server=NULL;
	} else if (snamelen==CS_NULLTERM) {
		server=(char *)servername;
	} else {
		server = (char *) malloc(snamelen+1);
		needfree++;
		strncpy(server,servername,snamelen);
		server[snamelen]='\0';
	}
        tds_set_server(con->tds_login,server);
        con->tds_socket = (void *) tds_connect(con->tds_login);

        tds_set_parent( con->tds_socket, con);

	if (needfree) free(server);

	return CS_SUCCEED;
}
CS_RETCODE ct_cmd_alloc(CS_CONNECTION *con, CS_COMMAND **cmd)
{

	*cmd = (CS_COMMAND *) malloc(sizeof(CS_COMMAND));
	memset(*cmd,'\0',sizeof(CS_COMMAND));

	/* so we know who we belong to */
	(*cmd)->con = con;

	return CS_SUCCEED;
}
CS_RETCODE ct_command(CS_COMMAND *cmd, CS_INT type, CS_VOID *buffer, CS_INT buflen, CS_INT option)
{
int query_len;

	/* FIX ME -- will only work for type CS_LANG_CMD */
	if (buflen==CS_NULLTERM) {
		query_len = strlen(buffer);
	} else {
		query_len = buflen;
	}
	if (cmd->query) free(cmd->query);
	cmd->query = (char *) malloc(query_len + 1);
	strncpy(cmd->query,(char *)buffer,query_len);
	cmd->query[query_len]='\0';

	return CS_SUCCEED;
}
CS_RETCODE ct_send(CS_COMMAND *cmd)
{
	tds_submit_query(cmd->con->tds_socket, cmd->query);
	return CS_SUCCEED;
}
CS_RETCODE ct_results(CS_COMMAND *cmd, CS_INT *result_type)
{
	if (cmd->cmd_done) {
		cmd->cmd_done = 0;
		*result_type = CS_CMD_DONE;
		return CS_SUCCEED;
	} 
	switch (tds_process_result_tokens(cmd->con->tds_socket)) {
		case TDS_SUCCEED:
			if (cmd->con->tds_socket->res_info->rows_exist) {
				*result_type = CS_ROW_RESULT;
				cmd->cmd_done = 1;
			} else {
				*result_type = CS_CMD_SUCCEED;
			}
			return CS_SUCCEED;
		case TDS_NO_MORE_RESULTS:
			*result_type = CS_CMD_DONE;
			return CS_END_RESULTS;
		default:
			return CS_FAIL;
	}	
	return CS_FAIL;
}
CS_RETCODE ct_bind(CS_COMMAND *cmd, CS_INT item, CS_DATAFMT *datafmt, CS_VOID *buffer, CS_INT *copied, CS_SMALLINT *indicator)
{
TDSCOLINFO * colinfo;
TDSRESULTINFO * resinfo;
TDSSOCKET * tds;

   tds = (TDSSOCKET *) cmd->con->tds_socket;
   resinfo = tds->res_info;
   colinfo = resinfo->columns[item-1];
   colinfo->varaddr = (char *)buffer;
   colinfo->column_bindtype = datafmt->datatype;
   colinfo->column_bindlen = datafmt->maxlength;

	return CS_SUCCEED;
}
CS_RETCODE ct_fetch(CS_COMMAND *cmd, CS_INT type, CS_INT offset, CS_INT option, CS_INT *rows_read)
{
	switch (tds_process_row_tokens(cmd->con->tds_socket)) {
		case TDS_SUCCEED:
			*rows_read = 1;
			_ct_bind_data(cmd);
			return TDS_SUCCEED;
		case TDS_NO_MORE_ROWS:
			*rows_read = 1;
			_ct_bind_data(cmd);
			return CS_END_DATA;
		default:
			*rows_read = 0;
			return CS_FAIL;
	}
	return CS_SUCCEED;
}
static void _ct_bind_data(CS_COMMAND *cmd)
{
int i;
TDSCOLINFO *curcol;
TDSRESULTINFO *resinfo = cmd->con->tds_socket->res_info;
unsigned char *src;
int srctype,desttype;

   for (i=0;i<resinfo->num_cols;i++)
   {
      curcol = resinfo->columns[i];
	/* currently there is no NULL handling for ctlib, this is where it 
	** will be */
	/* printf("%d %s\n",i,resinfo->columns[i]->column_value); */
		if (curcol->varaddr) {
			/* FIX ME -- I only work with strings so far */
			/* strcpy(curcol->varaddr,  */
			src = &(resinfo->current_row[curcol->column_offset]);
			srctype = curcol->column_type;
			desttype = _ct_get_server_type(curcol->column_bindtype);
			tds_convert(srctype,src, -1,
				desttype,curcol->varaddr,curcol->column_bindlen);
		}
	}
}
CS_RETCODE ct_cmd_drop(CS_COMMAND *cmd)
{
	if (cmd) free(cmd);
	return CS_SUCCEED;
}
CS_RETCODE ct_close(CS_CONNECTION *con, CS_INT option)
{
	close(con->tds_socket->s);
	return CS_SUCCEED;
}
CS_RETCODE ct_con_drop(CS_CONNECTION *con)
{
	if (con) free(con);
	return CS_SUCCEED;
}
int _ct_get_client_type(int datatype)
{
   switch (datatype) {
         case SYBCHAR:
         case SYBVARCHAR:
		return CS_CHAR_TYPE;
		break;
         case SYBINT4:
      		return CS_INT_TYPE;
		break;
         case SYBINT2:
      		return CS_SMALLINT_TYPE;
		break;
         case SYBINT1:
      		return CS_TINYINT_TYPE;
		break;
         case SYBREAL:
      		return CS_REAL_TYPE;
		break;
         case SYBFLT8:
      		return CS_FLOAT_TYPE;
		break;
         case SYBMONEY:
      		return CS_MONEY_TYPE;
		break;
         case SYBMONEY4:
      		return CS_MONEY4_TYPE;
		break;
         case SYBDATETIME:
      		return CS_DATETIME_TYPE;
		break;
         case SYBDATETIME4:
      		return CS_DATETIME4_TYPE;
		break;
   }
}
int _ct_get_server_type(int datatype)
{
   switch (datatype) {
      case CS_CHAR_TYPE:
         return SYBCHAR;
         break;
      case CS_INT_TYPE:
         return SYBINT4;
         break;
      case CS_SMALLINT_TYPE:
         return SYBINT2;
         break;
      case CS_TINYINT_TYPE:
         return SYBINT1;
         break;
      case CS_REAL_TYPE:
         return SYBREAL;
         break;
      case CS_FLOAT_TYPE:
         return SYBFLT8;
         break;
      case CS_MONEY_TYPE:
         return SYBMONEY;
         break;
      case CS_MONEY4_TYPE:
         return SYBMONEY4;
         break;
      case CS_DATETIME_TYPE:
         return SYBDATETIME;
         break;
      case CS_DATETIME4_TYPE:
         return SYBDATETIME4;
         break;
      default:
         return -1;
         break;
   }
}
CS_RETCODE ct_cancel(CS_CONNECTION *conn, CS_COMMAND *cmd, CS_INT type)
{
}
CS_RETCODE ct_describe(CS_COMMAND *cmd, CS_INT item, CS_DATAFMT *datafmt)
{
TDSRESULTINFO *resinfo = cmd->con->tds_socket->res_info;
TDSCOLINFO *curcol;

	if (item<1 || item>resinfo->num_cols) return CS_FAIL;	
	curcol=resinfo->columns[item-1];
	strncpy(datafmt->name, curcol->column_name, CS_MAX_NAME);
	datafmt->namelen = strlen(curcol->column_name);
	/* need to turn the SYBxxx into a CS_xxx_TYPE */
	datafmt->datatype = _ct_get_client_type(curcol->column_type); 
	datafmt->maxlength = curcol->column_size;
	datafmt->usertype = curcol->column_usertype; 
	datafmt->precision = curcol->column_prec; 
	datafmt->scale = curcol->column_scale; 
	/* we currently don't store this info.  default it for now */
	datafmt->status = 0;
	datafmt->count = 1;
	datafmt->locale = NULL;
	
	return CS_SUCCEED;
}
CS_RETCODE ct_res_info(CS_COMMAND *cmd, CS_INT type, CS_VOID *buffer, CS_INT buflen, CS_INT *out_len)
{
TDSRESULTINFO *resinfo = cmd->con->tds_socket->res_info;

	switch(type) {
		case CS_NUMDATA:
			memcpy(buffer, &(resinfo->num_cols), sizeof(int));
			break;
		case CS_ROW_COUNT:
			memcpy(buffer, &(resinfo->row_count), sizeof(long));
			break;
		default:
			fprintf(stderr,"Unknown type in ct_res_info: %d\n",type);
			return CS_FAIL;
			break;
	}
	return CS_SUCCEED;
}
CS_RETCODE ct_config(CS_CONTEXT *ctx, CS_INT action, CS_INT property, CS_VOID *buffer, CS_INT buflen, CS_INT *outlen)
{
	return CS_SUCCEED;
}
CS_RETCODE ct_cmd_props(CS_COMMAND *cmd, CS_INT action, CS_INT property, CS_VOID *buffer, CS_INT buflen, CS_INT *outlen)
{
	return CS_SUCCEED;
}
CS_RETCODE ct_compute_info(CS_COMMAND *cmd, CS_INT type, CS_INT colnum, CS_VOID *buffer, CS_INT buflen, CS_INT *outlen)
{
	return CS_SUCCEED;
}
CS_RETCODE ct_get_data(CS_COMMAND *cmd, CS_INT item, CS_VOID *buffer, CS_INT buflen, CS_INT *outlen)
{
	return CS_SUCCEED;
}
CS_RETCODE ct_capability(CS_CONNECTION *con, CS_INT action, CS_INT type, CS_INT capability, CS_VOID *value)
{
TDSLOGIN *login;
unsigned char *mask;
	
	login = (TDSLOGIN *) con->tds_login;
	mask = login->capabilities;

	if (action==CS_SET && type==CS_CAP_RESPONSE) {
		if (*((CS_BOOL *)value)==CS_TRUE) {
			switch(capability) {
			case CS_DATA_NOBOUNDARY:
				mask[13]|=0x01;break;
			case CS_DATA_NOTDSDEBUG:
				mask[13]|=0x02;break;
			case CS_RES_NOSTRIPBLANKS:
				mask[13]|=0x04;break;
			case CS_DATA_NOINT8:
				mask[13]|=0x08;break;
			case CS_DATA_NOINTN:
				mask[14]|=0x01;break;
			case CS_DATA_NODATETIMEN:
				mask[14]|=0x02;break;
			case CS_DATA_NOMONEYN:
				mask[14]|=0x04;break;
			case CS_CON_NOOOB:
				mask[14]|=0x08;break;
			case CS_CON_NOINBAND:
				mask[14]|=0x10;break;
			case CS_PROTO_NOTEXT:
				mask[14]|=0x20;break;
			case CS_PROTO_NOBULK:
				mask[14]|=0x40;break;
			case CS_DATA_NOSENSITIVITY:
				mask[14]|=0x80;break;
			case CS_DATA_NOFLT4:
				mask[15]|=0x01;break;
			case CS_DATA_NOFLT8:
				mask[15]|=0x02;break;
			case CS_DATA_NONUM:
				mask[15]|=0x04;break;
			case CS_DATA_NOTEXT:
				mask[15]|=0x08;break;
			case CS_DATA_NOIMAGE:
				mask[15]|=0x10;break;
			case CS_DATA_NODEC:
				mask[15]|=0x20;break;
			case CS_DATA_NOLCHAR:
				mask[15]|=0x40;break;
			case CS_DATA_NOLBIN:
				mask[15]|=0x80;break;
			case CS_DATA_NOCHAR:
				mask[16]|=0x01;break;
			case CS_DATA_NOVCHAR:
				mask[16]|=0x02;break;
			case CS_DATA_NOBIN:
				mask[16]|=0x04;break;
			case CS_DATA_NOVBIN:
				mask[16]|=0x08;break;
			case CS_DATA_NOMNY8:
				mask[16]|=0x10;break;
			case CS_DATA_NOMNY4:
				mask[16]|=0x20;break;
			case CS_DATA_NODATE8:
				mask[16]|=0x40;break;
			case CS_DATA_NODATE4:
				mask[16]|=0x80;break;
			case CS_RES_NOMSG:
				mask[17]|=0x02;break;
			case CS_RES_NOEED:
				mask[17]|=0x04;break;
			case CS_RES_NOPARAM:
				mask[17]|=0x08;break;
			case CS_DATA_NOINT1:
				mask[17]|=0x10;break;
			case CS_DATA_NOINT2:
				mask[17]|=0x20;break;
			case CS_DATA_NOINT4:
				mask[17]|=0x40;break;
			case CS_DATA_NOBIT:
				mask[17]|=0x80;break;
			}
		} else {
			switch(capability) {
			case CS_DATA_NOBOUNDARY:
				mask[13]&=(!0x01);break;
			case CS_DATA_NOTDSDEBUG:
				mask[13]&=(!0x02);break;
			case CS_RES_NOSTRIPBLANKS:
				mask[13]&=(!0x04);break;
			case CS_DATA_NOINT8:
				mask[13]&=(!0x08);break;
			case CS_DATA_NOINTN:
				mask[14]&=(!0x01);break;
			case CS_DATA_NODATETIMEN:
				mask[14]&=(!0x02);break;
			case CS_DATA_NOMONEYN:
				mask[14]&=(!0x04);break;
			case CS_CON_NOOOB:
				mask[14]&=(!0x08);break;
			case CS_CON_NOINBAND:
				mask[14]&=(!0x10);break;
			case CS_PROTO_NOTEXT:
				mask[14]&=(!0x20);break;
			case CS_PROTO_NOBULK:
				mask[14]&=(!0x40);break;
			case CS_DATA_NOSENSITIVITY:
				mask[14]&=(!0x80);break;
			case CS_DATA_NOFLT4:
				mask[15]&=(!0x01);break;
			case CS_DATA_NOFLT8:
				mask[15]&=(!0x02);break;
			case CS_DATA_NONUM:
				mask[15]&=(!0x04);break;
			case CS_DATA_NOTEXT:
				mask[15]&=(!0x08);break;
			case CS_DATA_NOIMAGE:
				mask[15]&=(!0x10);break;
			case CS_DATA_NODEC:
				mask[15]&=(!0x20);break;
			case CS_DATA_NOLCHAR:
				mask[15]&=(!0x40);break;
			case CS_DATA_NOLBIN:
				mask[15]&=(!0x80);break;
			case CS_DATA_NOCHAR:
				mask[16]&=(!0x01);break;
			case CS_DATA_NOVCHAR:
				mask[16]&=(!0x02);break;
			case CS_DATA_NOBIN:
				mask[16]&=(!0x04);break;
			case CS_DATA_NOVBIN:
				mask[16]&=(!0x08);break;
			case CS_DATA_NOMNY8:
				mask[16]&=(!0x10);break;
			case CS_DATA_NOMNY4:
				mask[16]&=(!0x20);break;
			case CS_DATA_NODATE8:
				mask[16]&=(!0x40);break;
			case CS_DATA_NODATE4:
				mask[16]&=(!0x80);break;
			case CS_RES_NOMSG:
				mask[17]&=(!0x02);break;
			case CS_RES_NOEED:
				mask[17]&=(!0x04);break;
			case CS_RES_NOPARAM:
				mask[17]&=(!0x08);break;
			case CS_DATA_NOINT1:
				mask[17]&=(!0x10);break;
			case CS_DATA_NOINT2:
				mask[17]&=(!0x20);break;
			case CS_DATA_NOINT4:
				mask[17]&=(!0x40);break;
			case CS_DATA_NOBIT:
				mask[17]&=(!0x80);break;
			}
		}
	} else if (action==CS_GET && type==CS_CAP_RESPONSE) {
		switch (capability) {
		case CS_DATA_NOBOUNDARY:
			*((CS_BOOL *)value)=mask[13]&0x01 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_NOTDSDEBUG:
			*((CS_BOOL *)value)=mask[13]&0x02 ? CS_TRUE : CS_FALSE;break;
		case CS_RES_NOSTRIPBLANKS:
			*((CS_BOOL *)value)=mask[13]&0x04 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_NOINT8:
			*((CS_BOOL *)value)=mask[13]&0x08 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_NOINTN:
			*((CS_BOOL *)value)=mask[14]&0x01 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_NODATETIMEN:
			*((CS_BOOL *)value)=mask[14]&0x02 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_NOMONEYN:
			*((CS_BOOL *)value)=mask[14]&0x04 ? CS_TRUE : CS_FALSE;break;
		case CS_CON_NOOOB:
			*((CS_BOOL *)value)=mask[14]&0x08 ? CS_TRUE : CS_FALSE;break;
		case CS_CON_NOINBAND:
			*((CS_BOOL *)value)=mask[14]&0x10 ? CS_TRUE : CS_FALSE;break;
		case CS_PROTO_NOTEXT:
			*((CS_BOOL *)value)=mask[14]&0x20 ? CS_TRUE : CS_FALSE;break;
		case CS_PROTO_NOBULK:
			*((CS_BOOL *)value)=mask[14]&0x40 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_NOSENSITIVITY:
			*((CS_BOOL *)value)=mask[14]&0x80 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_NOFLT4:
			*((CS_BOOL *)value)=mask[15]&0x01 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_NOFLT8:
			*((CS_BOOL *)value)=mask[15]&0x02 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_NONUM:
			*((CS_BOOL *)value)=mask[15]&0x04 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_NOTEXT:
			*((CS_BOOL *)value)=mask[15]&0x08 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_NOIMAGE:
			*((CS_BOOL *)value)=mask[15]&0x10 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_NODEC:
			*((CS_BOOL *)value)=mask[15]&0x20 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_NOLCHAR:
			*((CS_BOOL *)value)=mask[15]&0x40 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_NOLBIN:
			*((CS_BOOL *)value)=mask[15]&0x80 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_NOCHAR:
			*((CS_BOOL *)value)=mask[16]&0x01 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_NOVCHAR:
			*((CS_BOOL *)value)=mask[16]&0x02 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_NOBIN:
			*((CS_BOOL *)value)=mask[16]&0x04 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_NOVBIN:
			*((CS_BOOL *)value)=mask[16]&0x08 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_NOMNY8:
			*((CS_BOOL *)value)=mask[16]&0x10 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_NOMNY4:
			*((CS_BOOL *)value)=mask[16]&0x20 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_NODATE8:
			*((CS_BOOL *)value)=mask[16]&0x40 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_NODATE4:
			*((CS_BOOL *)value)=mask[16]&0x80 ? CS_TRUE : CS_FALSE;break;
		case CS_RES_NOMSG:
			*((CS_BOOL *)value)=mask[17]&0x02 ? CS_TRUE : CS_FALSE;break;
		case CS_RES_NOEED:
			*((CS_BOOL *)value)=mask[17]&0x04 ? CS_TRUE : CS_FALSE;break;
		case CS_RES_NOPARAM:
			*((CS_BOOL *)value)=mask[17]&0x08 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_NOINT1:
			*((CS_BOOL *)value)=mask[17]&0x10 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_NOINT2:
			*((CS_BOOL *)value)=mask[17]&0x20 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_NOINT4:
			*((CS_BOOL *)value)=mask[17]&0x40 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_NOBIT:
			*((CS_BOOL *)value)=mask[17]&0x80 ? CS_TRUE : CS_FALSE;break;
		}
	} else if (action==CS_GET && type==CS_CAP_REQUEST) {
		switch (capability) {
		case CS_PROTO_DYNPROC:
			*((CS_BOOL *)value)=mask[2]&0x01 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_FLTN:
			*((CS_BOOL *)value)=mask[2]&0x02 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_BITN:
			*((CS_BOOL *)value)=mask[2]&0x04 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_INT8:
			*((CS_BOOL *)value)=mask[2]&0x08 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_VOID:
			*((CS_BOOL *)value)=mask[2]&0x10 ? CS_TRUE : CS_FALSE;break;
		case CS_CON_INBAND:
			*((CS_BOOL *)value)=mask[3]&0x01 ? CS_TRUE : CS_FALSE;break;
		case CS_CON_LOGICAL:
			*((CS_BOOL *)value)=mask[3]&0x02 ? CS_TRUE : CS_FALSE;break;
		case CS_PROTO_TEXT:
			*((CS_BOOL *)value)=mask[3]&0x04 ? CS_TRUE : CS_FALSE;break;
		case CS_PROTO_BULK:
			*((CS_BOOL *)value)=mask[3]&0x08 ? CS_TRUE : CS_FALSE;break;
		case CS_REQ_URGNOTIF:
			*((CS_BOOL *)value)=mask[3]&0x10 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_SENSITIVITY:
			*((CS_BOOL *)value)=mask[3]&0x20 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_BOUNDARY:
			*((CS_BOOL *)value)=mask[3]&0x40 ? CS_TRUE : CS_FALSE;break;
		case CS_PROTO_DYNAMIC:
			*((CS_BOOL *)value)=mask[3]&0x80 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_MONEYN:
			*((CS_BOOL *)value)=mask[4]&0x01 ? CS_TRUE : CS_FALSE;break;
		case CS_CSR_PREV:
			*((CS_BOOL *)value)=mask[4]&0x02 ? CS_TRUE : CS_FALSE;break;
		case CS_CSR_FIRST:
			*((CS_BOOL *)value)=mask[4]&0x04 ? CS_TRUE : CS_FALSE;break;
		case CS_CSR_LAST:
			*((CS_BOOL *)value)=mask[4]&0x08 ? CS_TRUE : CS_FALSE;break;
		case CS_CSR_ABS:
			*((CS_BOOL *)value)=mask[4]&0x10 ? CS_TRUE : CS_FALSE;break;
		case CS_CSR_REL:
			*((CS_BOOL *)value)=mask[4]&0x20 ? CS_TRUE : CS_FALSE;break;
		case CS_CSR_MULTI:
			*((CS_BOOL *)value)=mask[4]&0x40 ? CS_TRUE : CS_FALSE;break;
		case CS_CON_OOB:
			*((CS_BOOL *)value)=mask[4]&0x80 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_NUM:
			*((CS_BOOL *)value)=mask[5]&0x01 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_TEXT:
			*((CS_BOOL *)value)=mask[5]&0x02 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_IMAGE:
			*((CS_BOOL *)value)=mask[5]&0x04 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_DEC:
			*((CS_BOOL *)value)=mask[5]&0x08 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_LCHAR:
			*((CS_BOOL *)value)=mask[5]&0x10 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_LBIN:
			*((CS_BOOL *)value)=mask[5]&0x20 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_INTN:
			*((CS_BOOL *)value)=mask[5]&0x40 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_DATETIMEN:
			*((CS_BOOL *)value)=mask[5]&0x80 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_BIN:
			*((CS_BOOL *)value)=mask[6]&0x01 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_VBIN:
			*((CS_BOOL *)value)=mask[6]&0x02 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_MNY8:
			*((CS_BOOL *)value)=mask[6]&0x04 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_MNY4:
			*((CS_BOOL *)value)=mask[6]&0x08 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_DATE8:
			*((CS_BOOL *)value)=mask[6]&0x10 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_DATE4:
			*((CS_BOOL *)value)=mask[6]&0x20 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_FLT4:
			*((CS_BOOL *)value)=mask[6]&0x40 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_FLT8:
			*((CS_BOOL *)value)=mask[6]&0x80 ? CS_TRUE : CS_FALSE;break;
		case CS_REQ_MSG:
			*((CS_BOOL *)value)=mask[7]&0x01 ? CS_TRUE : CS_FALSE;break;
		case CS_REQ_PARAM:
			*((CS_BOOL *)value)=mask[7]&0x02 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_INT1:
			*((CS_BOOL *)value)=mask[7]&0x04 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_INT2:
			*((CS_BOOL *)value)=mask[7]&0x08 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_INT4:
			*((CS_BOOL *)value)=mask[7]&0x10 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_BIT:
			*((CS_BOOL *)value)=mask[7]&0x20 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_CHAR:
			*((CS_BOOL *)value)=mask[7]&0x40 ? CS_TRUE : CS_FALSE;break;
		case CS_DATA_VCHAR:
			*((CS_BOOL *)value)=mask[7]&0x80 ? CS_TRUE : CS_FALSE;break;
		case CS_REQ_LANG:
			*((CS_BOOL *)value)=mask[8]&0x02 ? CS_TRUE : CS_FALSE;break;
		case CS_REQ_RPC:
			*((CS_BOOL *)value)=mask[8]&0x04 ? CS_TRUE : CS_FALSE;break;
		case CS_REQ_NOTIF:
			*((CS_BOOL *)value)=mask[8]&0x08 ? CS_TRUE : CS_FALSE;break;
		case CS_REQ_MSTMT:
			*((CS_BOOL *)value)=mask[8]&0x10 ? CS_TRUE : CS_FALSE;break;
		case CS_REQ_BCP:
			*((CS_BOOL *)value)=mask[8]&0x20 ? CS_TRUE : CS_FALSE;break;
		case CS_REQ_CURSOR:
			*((CS_BOOL *)value)=mask[8]&0x40 ? CS_TRUE : CS_FALSE;break;
		case CS_REQ_DYN:
			*((CS_BOOL *)value)=mask[8]&0x80 ? CS_TRUE : CS_FALSE;break;
		}
	} else {
		/* bad values */
		return CS_FAIL;
	}
	return CS_SUCCEED;
}
CS_RETCODE ct_dynamic(CS_COMMAND *cmd, CS_INT type, CS_CHAR *id, CS_INT idlen, CS_CHAR *buffer, CS_INT buflen)
{
	return CS_SUCCEED;
}
CS_RETCODE ct_param(CS_COMMAND *cmd, CS_DATAFMT *datafmt, CS_VOID *data, CS_INT datalen, CS_SMALLINT indicator)
{
}
